/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#ifndef __DRAWING_OBJECT_H__
#define __DRAWING_OBJECT_H__

class Context;
class CTinyCadDoc;

enum paint_options
{
	draw_normal,
	draw_selected,
	draw_selectable,
};

#include "Stream.h"


// The different types of object
enum ObjType {
	xEndFile = 0, xNoDraw, xAnotation, xDesignInformation, xIndexInformation, xSymbol,
	xError, xFont, xSymbolEx, xLayerOld, xLineStyle, xFillStyle, xLayer, xLayerTable, 
	xSymbolEx2,

	xBusName = 15,
	xNoConnect = 16, xJunction, xLabel, xWire, xBus, xMethod,
	xPower, xPin, xBusSlash, xLabelEx, xMethodEx, xMethodEx2, xPinEx,

	xLine = 32, xDash, xText, xCircle, xSquare, xArc, xLineEx, xTextEx,
	xCircleEx, xSquareEx, xArcEx,

	xLineEx2, xCircleEx2, xSquareEx2, xArcEx2, xMetaFile, 

	xTag  = 48, xRuler, xMethodEx3, xPolygon,

	xEditItem = 128, 
	xDesignInformation2,
	xMetaFiles,
	xOptions,
	xSymbols, xSymbolMethod, xSymbolDelete,
	xCircleEx3, xSquareEx3, xTextEx2, xLabelEx2, xBusNameEx
};


//////// The is the main object that all objects inherit from //////

class CDrawingObject 
{

public:

	CTinyCadDoc	*m_pDesign;					// Our current design that this object belongs to

	CPoint a;								// position of this object (in internal units)
	CPoint b;								// the size of the object, these two points define
											// the bounding box of the object

	char m_segment;							// Mode of current edit
	int	 m_Undo_Level;						// The timestamp when this object was created
	int	 m_Delete_Level;

	BOOL	DoesExist();		// Has this object been undone or deleted?

	CDrawingObject( CTinyCadDoc *pDesign );
	virtual ~CDrawingObject() { }			// The destructor


	CPoint 	ReadPoint(CStream &);
	void 	WritePoint(CStream &, CPoint);


	virtual void Save(CStream &);			// Save data into file
	virtual void Load(CStream &);			// Load data into this object

	virtual double DistanceFromPoint( CPoint p );
	virtual BOOL IsInside(int left,int right,int top,int bottom);	// Is this object inside this rectangle?
	virtual BOOL CanEdit();					// Is this object editable after placed?
	virtual CString GetName() const;		// Get the string name of this object
	virtual ObjType GetType();				// Get this object's type

	virtual void TagResources();			// Tag the resources this object is currently using
	virtual CString Find(const char *);		// Find a string in this object
	virtual	void NewOptions();		// Indicate to an object being edited that the options have changed

	virtual void BeginEdit(BOOL re_edit);				// Called before starting editing this object
	virtual void EndEdit();		// Called before deleting this object from current edit

	virtual void Display( BOOL erase = TRUE );
	virtual void Paint(Context &, paint_options  );	// Display this object on the screen
	virtual void PaintHandles(Context & );			// Display this object's handles (for editing)
	
	virtual void Move(CPoint);						// Follow a mouse movement
	virtual void LButtonDown(CPoint,CPoint);		// The user has clicked with the left hand button
	virtual void DblLButtonDown(CPoint,CPoint);		// The user has double clicked with the left hand button
	virtual BOOL RButtonDown(CPoint,CPoint);		// The user has clicked with the right hand button
	virtual void LButtonUp(CPoint);					// The user has released the left hand button
	virtual void RButtonUp(CPoint,CPoint);			// The user has released the right hand button

	// For the context menu
	virtual void ContextMenu( CPoint p, UINT id );
	virtual void FinishDrawing( CPoint p );
	virtual int  GetContextMenu();

	
	// Move fields of this object about
	virtual int IsInsideField(CPoint p);
	virtual void MoveField(int w, CPoint r);
	virtual int SetCursor( CPoint p );
	virtual int SetCursorEdit( CPoint p );


	virtual CDrawingObject* Store();		// Store current object
	virtual void Rotate(CPoint,int);		// Rotate this item about a point

	virtual BOOL IsConstruction();

	virtual int getMenuID() { return -1; }
};


typedef std::list<CDrawingObject*> drawingCollection;
typedef drawingCollection::iterator drawingIterator;


#endif
