/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002,2003 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "stdafx.h"
#include <set>
#include <map>

#include "TinyCadView.h"
#include "TinyCad.h"
#include "special.h"
#include "library.h"
#include "net.h"
#include "registry.h"
#include "diag.h"
extern CDlgERCListBox theERCListBox;



////// These are the operators on B-Trees //////

// Add a node to the node tree
void BTree::DoAdd(BTree *ins)
{

  // Insert into the tree in an iterative manner
  BTree *find = this, *last = this;

  while (find!=NULL) {
	last = find;
	if (find->key < ins->key)
		find=find->left;
	else
		find=find->right;
  }


  if (last->key < ins->key)
	last->left = ins;
  else
	last->right = ins;
}





////// These are the operators on the current NetList //////

// The constructor for a net
Net::Net(int NewValue,Node *NewFirst)
{
  value = NewValue;
  first = NewFirst;
  next = NULL;
}

// The destructor
Net::~Net()
{
  if (first!=NULL)
	delete first;

  if (next != NULL)
	delete next;
}


////// The hash table methods //////

// The constructor
NetHash::NetHash(int NewSize)
{
  size = NewSize;
  CurrentNet = 0;

  array = new Net*[size];

  for (int lp =0;lp!=size;lp++)
	array[lp] = NULL;
}


// The destructor
NetHash::~NetHash()
{
  for (int lp = 0;lp!=size;lp++)
	if (array[lp]!=NULL)
		delete array[lp];
}


Node *NetHash::GetList(int list)
{
  Net *n = array[list % size];

  while (n!=NULL && n->value!=list)
	n = n->next;

  return n==NULL ? NULL : n->first;
}


void NetHash::SetList(int list, Node *nd)
{
  Net *n = array[list % size];

  while (n!=NULL && n->value!=list)
	n = n->next;


  if (n == NULL) {
	Net *nNet = new Net(list,nd);
	nNet->next = array[list % size];
	array[list % size] = nNet;
  } else {
	n->first = nd;
  }

}





////// These are the operators on the Net Nodes //////

// Create a node
Node::Node(CDrawingObject *NewParent,CPoint NewA,CString NewName)
{
  a = NewA;
  name = NewName;
  key = a.x * a.y;
  parent = NewParent;
  NetList = -1;
  NetPtr = NULL;
  left = NULL;
  right = NULL;
}


// Delete a node
Node::~Node()
{
  if (NetPtr != NULL)
	delete NetPtr;
}




// Add a node to the tree (if necessary), return value of the netlist
int Node::Add(NetHash *theNetHash,Node *ins)
{

  // Is this node already in the tree?
  Node *found = Find(ins);

  // If the found item is the same as this, then there is no need to store it
  if (found != NULL
   && found->name == ins->name 
   && found->NetList == ins->NetList) {
	// Free the memory used by this netlist
	delete ins;
	return found->NetList;
  }
  
  if (ins->NetList == -1) {
	if (found!=NULL) {
		if (found->name != ins->name) {
			ins->NetList = found->NetList;
			ins->NetPtr = theNetHash->GetList(ins->NetList);
			theNetHash->SetList(ins->NetList,ins);
			DoAdd(ins);
		}
		return found->NetList;
	} else {
		// Create a new net list
		ins->NetList = theNetHash->GetNewNet();
		theNetHash->SetList(ins->NetList,ins);
		DoAdd(ins);

		return ins->NetList;
	}
  } else {
	if (found!=NULL && found->NetList != ins->NetList) {
		// The two nets must be joined
		int NewList = ins->NetList;
		int OldList = found->NetList;
		Node *pointer = theNetHash->GetList(OldList);
		Node *last = NULL;
		while (pointer!=NULL) {
			pointer->NetList = NewList;
			last = pointer;
			pointer=pointer->NetPtr;
		}
		// Now join the two lists
		if (last != NULL)
			last->NetPtr = theNetHash->GetList(NewList);

		// Delete one of the lists
		theNetHash->SetList(NewList,theNetHash->GetList(OldList));
		theNetHash->SetList(OldList, NULL);
		ins->NetList = NewList;
	}
	if (found==NULL || ins->name != found->name) {
		ins->NetPtr = theNetHash->GetList(ins->NetList);
		theNetHash->SetList(ins->NetList,ins);
		DoAdd(ins);

		return ins->NetList;
	} else {
		// If this item is not stored then it's memory must be freed
		int hold = ins->NetList;
		delete ins;

		return hold;
	}
  }
}



Node *Node::Find(Node *fnd)
{
  // Is this the node being searched for?
  Node *find = this;

  while (find != NULL && find->a != fnd->a) {
	if (find->key < (fnd->key))
		// Tranverse the left branch
		find = (Node *)(find->left);
	else
		// Tranverse the right branch
		find = (Node *)(find->right);
  }

  return find;
}



////// Generate a netlist for a file //////


// Connections between nets (the nodes) are at:
// .. the end of wires
// .. the point of pins
// .. the point of power
// .. the point of labels
// .. the point of no-connects

// To generate a netlist:
// 1. Search design for nodes.
//  - Insert nodes into node tree and either assign net list to node or gernerate a new net list
//  - If wire joins two different netlists, then join the two netlists together
// 2. Find Junctions
//  - Find which wires junctions lie on and connect their netlists
// 3. Find Labels
//  - Find which wire each label lies on and connect their netlists



int TheAdd(NetHash *theNetHash,Node *&nodes,Node *n)
{
  int r;

  if (nodes==NULL) {
	n->NetList = theNetHash->GetNewNet();
	theNetHash->SetList(n->NetList,n);
	nodes = n;
	r = n->NetList;
  } else
	r = nodes->Add(theNetHash,n);

  return r;
}





NetHash *MakeNet( CTinyCadDoc *pDesign )
{
  Node *nodes=NULL;
  NetHash *theNetHash = new NetHash(NetHashSize);
  Node *n;

  typedef std::map<CString,int> string_map;
  string_map	Powers;
  string_map	Connected;

  // Search for nodes, and build the node tree
  drawingIterator it = pDesign->GetDrawingBegin();
  while (it != pDesign->GetDrawingEnd()) 
  {
	CDrawingObject *ObjPtr = *it;
	string_map::iterator found;
	int hold;
	CPoint tr;

	switch (ObjPtr->GetType()) 
	{
		case xMethodEx3:
			#define thePin ((CDrawPin*)pointer)
			#define theMethod ((CDrawMethod*)ObjPtr)
			hold = pDesign->GetPart();
			pDesign->SetPart(theMethod->GetSubPart());

			{
				drawingCollection method;
				((CDrawMethod *)ObjPtr)->ExtractSymbol(tr,method);

				drawingIterator it = method.begin();
				while ( it != method.end() ) 
				{
					CDrawingObject *pointer = *it;

					if (pointer->GetType()==xPinEx && !(thePin->IsPower()) && !(thePin->IsInvisible())) 
					{
						n = new Node(thePin,thePin->GetPosition(theMethod),
							"(" +theMethod->GetRef()+","+thePin->GetNumber()+")");
						n->pMethod = theMethod;
						TheAdd(theNetHash,nodes,n);
					}

					++ it;
				}
				pDesign->SetPart(hold);

				// Has this symbol had it's power connected?
				if (Connected.find(theMethod->GetRef()) == Connected.end()) 
				{
					Connected[ theMethod->GetRef() ] = TRUE;

					drawingCollection method;
					((CDrawMethod *)ObjPtr)->ExtractSymbol(tr,method);
					drawingIterator it = method.begin();
					while (it!=method.end()) 
					{
						CDrawingObject *pointer = *it;

						if (pointer->GetType()==xPinEx && thePin->IsPower()) 
						{
							n = new Node(thePin,thePin->GetPosition(theMethod),
								"(" +theMethod->GetRef()+","+thePin->GetNumber()+")");
							n->pMethod = theMethod;

							// Look up the netlist this power belongs to
							found = Powers.find( thePin->GetPinName() );
							if (found != Powers.end())
								n->NetList = (*found).second;
							hold = TheAdd(theNetHash,nodes,n);
							if (found == Powers.end())
								Powers[thePin->GetPinName()] = hold;
						}

						++ it;
					}
				}
			}			
			break;
		case xNoConnect:
			n = new Node(ObjPtr,ObjPtr->a,"");
			TheAdd(theNetHash,nodes,n);			
			break;
		case xJunction:
			n = new Node(ObjPtr,ObjPtr->a,"");
			TheAdd(theNetHash,nodes,n);
			break;			
		case xPower:
			n = new Node(ObjPtr,ObjPtr->a,((CDrawPower *)ObjPtr)->GetValue());
			// Does this power item exist?
			found = Powers.find(((CDrawPower *)ObjPtr)->GetValue());
			if (found != Powers.end())
				n->NetList = (*found).second;
			hold = TheAdd(theNetHash,nodes,n);
			if (found == NULL)
				Powers[((CDrawPower *)ObjPtr)->GetValue()] = hold;
			break;
		case xWire:
			n = new Node(ObjPtr,ObjPtr->a,"");
			hold = TheAdd(theNetHash,nodes,n);
			n = new Node(ObjPtr,ObjPtr->b,"");
			n->NetList = hold;
			TheAdd(theNetHash,nodes,n);
			break;
	}
	
	++ it;
  }

 

  // Search for junctions and connect together
  it = pDesign->GetDrawingBegin();
  while (it != pDesign->GetDrawingEnd()) 
  {
	CDrawingObject *ObjPtr = *it;

	// Search for junctions
	if (ObjPtr->GetType() == xJunction) {
		// Find out which netlist was assigned to this point
		CPoint a = ObjPtr->a;
		n = new Node(NULL,a,"");
		int NetNumber = (nodes->Find(n))->NetList;
		delete n;

		// Look for wires which cross this junction
		// If the wire has an end at this junction then it is already connected
		drawingIterator search_it = pDesign->GetDrawingBegin();
		while (search_it != pDesign->GetDrawingEnd()) 
		{
			CDrawingObject *search = *search_it;

			if (search->GetType()==xWire 
			 && search->a!=a && search->b!=a
			 && search->IsInside(a.x,a.x,a.y,a.y)) 
			{
				n = new Node(NULL,search->a,"");
				n->NetList = NetNumber;
				NetNumber = TheAdd(theNetHash,nodes,n);
			}

			++ search_it;
		}
	}
	
	++ it;
  }

  // Search for labels and connect to their respective lines
  string_map labels;
  it = pDesign->GetDrawingBegin();
  while (it != pDesign->GetDrawingEnd()) 
  {
	CDrawingObject *ObjPtr = *it;

	// Search for junctions
	if (ObjPtr->GetType() == xLabelEx2) 
	{
		CPoint a = ObjPtr->a;
		// Search for a wire this label is connect to
		// Only atempt to connect to a single wire
		drawingIterator search_it = pDesign->GetDrawingBegin();
		while (it != pDesign->GetDrawingEnd()) 
		{
			CDrawingObject *search = *search_it;
			if (search->GetType()==xWire && search->IsInside(a.x,a.x,a.y,a.y)) 
			{
				a = search->a;
				break;
			}
			
			++ search_it;
		}
		// Look up this label

		n = new Node(ObjPtr,a,((CDrawText *)ObjPtr)->GetValue());
		string_map::iterator found = labels.find(((CDrawText *)ObjPtr)->GetValue());
		if (found!=labels.end()) {
			n->NetList = (*found).second;
			n->name = "";
		}

		int hold = TheAdd(theNetHash,nodes,n);

		if (found == labels.end())
			labels[ ((CDrawText *)ObjPtr)->GetValue()] = hold;
	}
	
	++ it;
  }  



  return theNetHash;
}





////// Generate the net list for this design //////


void CTinyCadView::OnSpecialNet()
{
  // Get rid of any drawing tool
  GetDocument()->SelectObject(new CDrawEditItem(GetDocument()));

  // Get the file in which to save the network
  OPENFILENAME ofn;
  char szFile[256], szFileTitle[256];

  strcpy(szFile,GetDocument()->GetPathName());
  for (int i=0;szFile[i]!=0 && szFile[i]!='.';i++) ;
  strcpy(szFile+i, ".net");

  strcpy(szFileTitle,szFile);

  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = this->m_hWnd;
  ofn.lpstrFilter = "Netlist (*.net)\0*.net\0All files (*.*)\0*.*\0\0";
  ofn.lpstrCustomFilter = NULL;
  ofn.nMaxCustFilter = 0;
  ofn.nFilterIndex = 1;
  ofn.lpstrFile = szFile;
  ofn.nMaxFile = sizeof(szFile);
  ofn.lpstrFileTitle = szFileTitle;
  ofn.nMaxFileTitle = sizeof(szFileTitle);
  ofn.lpstrInitialDir = NULL;
  ofn.lpstrTitle = NULL;
  ofn.Flags = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST  | OFN_NOCHANGEDIR;
  ofn.nFileOffset = 0;
  ofn.nFileExtension = 0;
  ofn.lpstrDefExt = "net";
  ofn.lpstrTitle = "Create Net List";

  if (!GetSaveFileName(&ofn))
      	return;


  FILE *theFile = fopen(szFile,"w");
  if (theFile == NULL) {
	Message(IDS_CANNOTOPEN);
	return;
  }

  // Set the Busy icon
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_WAIT ) );

  // Get the net list
  NetHash *theNetHash = MakeNet( GetDocument() );

  fprintf(theFile,NetComment" ====+  Net List for %s  +====\n\n",(const char *)(GetDocument()->GetPathName()));


  fprintf(theFile,NetComment" ======+ The component list\n\n");
  {

  CString last = "";
  drawingIterator it = GetDocument()->GetDrawingBegin();
  while (it != GetDocument()->GetDrawingEnd()) 
  {
	CDrawingObject *pointer = *it;

	if (pointer->GetType() == xMethodEx3 && ((CDrawMethod*)pointer)->GetRef()!=last) 
	{
		const char *Name = (const char *)(((CDrawMethod *)pointer)->GetField(CDrawMethod::Name));
		const char *Ref  = (const char *)(((CDrawMethod *)pointer)->GetRef());
		fprintf(theFile,"COMPONENT '%s' = %s\n",Ref,Name);
		last = ((CDrawMethod *)pointer)->GetRef();
	}
	
	++ it;
  }
  }

  fprintf(theFile,"\n\n"NetComment" ======+ The net list\n\n");
  int Label = 0;
  for (int lp = 0;lp!=NetHashSize;lp++) {
	Net *theNet = (theNetHash->array)[lp];
	while (theNet != NULL) {
		Node *theNode = theNet->first;
		CString theLine,theLabel;

		if (theNode!=NULL) {
			theLine = "";
			BOOL first = TRUE,PrintLine=FALSE, Labeled = FALSE;
			while (theNode != NULL) {
				if (theNode->name!="") {
					if (Labeled || (theNode->parent)->GetType() != xPower && (theNode->parent)->GetType() != xLabelEx2) {
						if (first)
							first = FALSE;
						else
							theLine += ",";
						theLine += theNode->name;
						PrintLine=TRUE;
					} else {
						theLabel = theNode->name;
						Labeled = TRUE;
					}
				}
				theNode = theNode->NetPtr;
			}
			if (PrintLine) {
				fprintf(theFile,"NET  ");
				if (Labeled)
					fprintf(theFile,"'%s'",(const char *)theLabel);
				else
					fprintf(theFile,"'N%06d'",Label++);
				fprintf(theFile," =  %s\n",(const char *)theLine);
			}
		}
		theNet = theNet->next;
	}
  }

  delete theNetHash;

  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_ARROW ) );
  fclose(theFile);
}


////// Check the design rules for this design //////


////// The ERCBox dialog //////


BOOL CDlgERCBox::OnInitDialog()
{

  // Copy the errortest into the dialog
  CheckDlgButton(ERC_DUPREF,theErrorTest.DupRef);
  CheckDlgButton(ERC_UNCONNECT,theErrorTest.UnConnect);
  CheckDlgButton(ERC_NOCONNECT,theErrorTest.NoConnect);

  CheckDlgButton(ERC_POWER,theErrorTest.Power);
  CheckDlgButton(ERC_OUTPUTTOPWR,theErrorTest.OutputPwr);
  CheckDlgButton(ERC_OUTPUT,theErrorTest.Output);
  CheckDlgButton(ERC_NOUTPUT,theErrorTest.NoOutput);
  CheckDlgButton(ERC_UNCONNECTED,theErrorTest.UnConnected);

  return TRUE;
}


void CDlgERCBox::OnOK()
{

  // Copy the dialog into the errortest
  theErrorTest.DupRef		=IsDlgButtonChecked(ERC_DUPREF)!=0;
  theErrorTest.UnConnect	=IsDlgButtonChecked(ERC_UNCONNECT)!=0;
  theErrorTest.NoConnect	=IsDlgButtonChecked(ERC_NOCONNECT)!=0;

  theErrorTest.Power		=IsDlgButtonChecked(ERC_POWER)!=0;
  theErrorTest.OutputPwr	=IsDlgButtonChecked(ERC_OUTPUTTOPWR)!=0;
  theErrorTest.Output		=IsDlgButtonChecked(ERC_OUTPUT)!=0;
  theErrorTest.NoOutput		=IsDlgButtonChecked(ERC_NOUTPUT)!=0;
  theErrorTest.UnConnected	=IsDlgButtonChecked(ERC_UNCONNECTED)!=0;

  EndDialog( IDOK );
}




////// the menu entry point for this special function //////


// This table determines the type of the netlist

const int ErcTable[8][6] = {
// Next >	Power, NoConnect, Input, Output, Passive, TriState/BiDir

  nPower, nNoConnectOnly, nInput, nOutput, nPassive, nBiDir,					// Unknown
  nPower, nCheck1, nPassive, nOutput, nPassive, nBiDir,						// Passive
  nPower, nCheck1, nInput, nOutput, nPassive, nBiDir,						// Input
  ERR_POWERBIDIR, nCheck1, nOutput, ERR_OUTPUT, nOutput, ERR_OUTPUTBIDIR,			// Output
  ERR_POWERBIDIR, nCheck1, nBiDir, ERR_OUTPUTBIDIR, nBiDir, nBiDir,				// BiDir
  nPower, nCheck1, nPower, ERR_OUTPUTTOPWR, nPower, ERR_POWERBIDIR,				// Power
  ERR_NOCONNECT, ERR_NOCONNECT, ERR_NOCONNECT, ERR_NOCONNECT, ERR_NOCONNECT, ERR_NOCONNECT, 	// NoConnect
  nNoConnect, ERR_NOCONNECT, nNoConnect, nNoConnect, nNoConnect,nNoConnect			// NoConnectOnly
};

void CTinyCadView::OnSpecialCheck()
{

  // Get rid of any drawing tool
  GetDocument()->SelectObject(new CDrawEditItem(GetDocument()));

  CDlgERCBox theDialog;
  Registry theRegistry;

  static union { ErrorTest e; WORD i; } theErrorTest;

  WORD hold;
  theErrorTest.i = theRegistry.Get("ERC",&hold) ? hold : 0xffff;


  theDialog.SetErrorTest(theErrorTest.e);
  if (theDialog.DoModal()!=IDOK)
	return;

  theErrorTest.e = theDialog.GetErrorTest();  
  theRegistry.Place("ERC",theErrorTest.i);

  // Set the Busy icon
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_WAIT ) );

  // Generate the netlist
  NetHash *theNetHash = MakeNet( GetDocument() );

  // Delete all the errors which are currently in the design
  theERCListBox.Close();
  GetDocument()->DeleteErrors();
  theERCListBox.Open(GetDocument());
  int CurrentError = 0;

  // Scan the design for duplicated references
  if ((theErrorTest.e).DupRef)
  {
	  CString last = "";
	  drawingIterator it = GetDocument()->GetDrawingBegin();
	  while (it != GetDocument()->GetDrawingEnd()) 
	  {
		CDrawingObject *pointer = *it;

		if (pointer->GetType()==xMethodEx3) 
		{
			if (last == ((CDrawMethod *)pointer)->GetField(CDrawMethod::Ref)) {
				char buffer[256];
				if (LoadString(AfxGetInstanceHandle(),ERR_DUPREF,buffer,sizeof(buffer))==0)
					Message(IDS_NORES,MB_ICONEXCLAMATION | MB_OK);
				else {
					GetDocument()->Add(new CDrawError(GetDocument(),((CDrawMethod *)pointer)->GetFieldPos(CDrawMethod::Ref),CurrentError++));
					theERCListBox.AddString(buffer);
				}

			}
			last = ((CDrawMethod *)pointer)->GetField(CDrawMethod::Ref);
		}
		
		++ it;
	  }
  }


  // Scan netlist to determine the type of each net
  for (int lp = 0;lp!=NetHashSize;lp++) {
	Net *theNet = (theNetHash->array)[lp];
	while (theNet != NULL) {
		Node *theNode = theNet->first;
		if (theNode!=NULL) {
			int theNetType = nUnknown;
			Node *last = theNode;
			CString lastPower = "";
			int connections = 0;
			while (theNetType < ERR_BASE && theNode != NULL) {
				#define pObject (theNode->parent)
				#define pPin ((CDrawPin *)pObject)
				if (pObject != NULL) {
					// Determine the type of this node
					last = theNode;
					int t = -1;
					switch (pObject->GetType()) {
					case xPower:
						t = 0;
						if (lastPower=="")
							lastPower = ((CDrawPower *)pObject)->GetValue();
						else {
							if (lastPower!=(((CDrawPower *)pObject)->GetValue())) {
								theNetType = ERR_POWER;
								t = -1;
							}
						}
						connections ++;
						break;
					case xNoConnect:
						t = 1;
						connections ++;
						break;
					case xPin:
					case xPinEx:
						switch(pPin->GetElec()) {
						case 0: t =2; break;
						case 1: t =3; break;
						case 2: t =5; break;
						case 3: t =5; break;
						case 4: t =4; break;
						}
						if (pPin->IsPower()) {
							t = -1;
							theNetType = nPower;
						}
						connections ++;
						break;
					}
					theNetType = (t!=-1) ? ErcTable[theNetType][t] : theNetType;
					if (theNetType == nCheck1) {
						if (connections == 1)
							theNetType = nNoConnect;
						else
							theNetType = ERR_UNCONNECT;
					}
				}
				theNode = theNode->NetPtr;
			}
			theNode = last;
			CPoint pos = pObject->GetType()==xPinEx ? pPin->GetPosition(theNode->pMethod) : pObject->a;
			int ErrorNumber;
			if (connections == 1)
				theNetType = ERR_UNCONNECT;

			switch (theNetType) {
				case nUnknown:
					ErrorNumber = ERR_UNCONNECTED;
					break;
				case nInput:
					ErrorNumber = ERR_NOUTPUT;
					break;
				default:
					ErrorNumber = theNetType;
					break;
			}
			// Is this error to be reported?
			switch (ErrorNumber) {
			case ERR_UNCONNECT:
				if (!((theErrorTest.e).UnConnect))
					ErrorNumber = -1;
				break;
			case ERR_POWER:
				if (!((theErrorTest.e).Power))
					ErrorNumber = -1;
				break;
			case ERR_NOCONNECT:
				if (!((theErrorTest.e).NoConnect))
					ErrorNumber = -1;
				break;
			case ERR_NOUTPUT:
				if (!((theErrorTest.e).NoOutput))
					ErrorNumber = -1;
				break;
			case ERR_DUPREF:
				if (!((theErrorTest.e).DupRef))
					ErrorNumber = -1;
				break;
			case ERR_OUTPUT:
				if (!((theErrorTest.e).Output))
					ErrorNumber = -1;
				break;
			case ERR_OUTPUTTOPWR:
				if (!((theErrorTest.e).OutputPwr))
					ErrorNumber = -1;
				break;
			case ERR_UNCONNECTED:
				if (!((theErrorTest.e).UnConnected))
					ErrorNumber = -1;
				break;
			case ERR_OUTPUTBIDIR:
				if (!((theErrorTest.e).Output))
					ErrorNumber = -1;
				break;
			case ERR_POWERBIDIR:
				if (!((theErrorTest.e).OutputPwr))
					ErrorNumber = -1;
				break;
			}

			if (ErrorNumber >= ERR_BASE) {
				char buffer[256];
				if (LoadString(AfxGetInstanceHandle(),ErrorNumber,buffer,sizeof(buffer))==0)
					Message(IDS_NORES,MB_ICONEXCLAMATION | MB_OK);
				else {
					GetDocument()->Add(new CDrawError(GetDocument(),pos,CurrentError++));
					theERCListBox.AddString(buffer);
				}
			}
		}
		theNet = theNet->next;
	}
  }

  // Were any errors detected?
  if (CurrentError == 0) {
	char buffer[256];
	if (LoadString(AfxGetInstanceHandle(),ERR_NOERROR,buffer,sizeof(buffer))==0)
		Message(IDS_NORES,MB_ICONEXCLAMATION | MB_OK);
	else
		theERCListBox.AddString(buffer);
  }

  // Get rid of the netlist
  delete theNetHash;

  // Set the normal icon
  SetCursor( AfxGetApp()->LoadStandardCursor( IDC_ARROW ) );

  // Re-Draw the window
  Invalidate();

}

